@@ -44,24 +44,29 @@
 
 typedef struct _PangoFT2FontClass   PangoFT2FontClass;
 typedef struct _PangoFT2MetricsInfo PangoFT2MetricsInfo;
-typedef struct _PangoFT2ContextInfo PangoFT2ContextInfo;
 
 struct _PangoFT2FontClass
 {
   PangoFontClass parent_class;
 };
 
+struct _PangoFT2MetricsInfo
+{
+  const char       *sample_str;
+  PangoFontMetrics *metrics;
+};
+
 typedef struct
 {
   FT_Bitmap bitmap;
   int bitmap_left;
   int bitmap_top;
-}  PangoFT2RenderedGlyph;
+} PangoFT2RenderedGlyph;
 
 static PangoFontClass *parent_class;	/* Parent class structure for PangoFT2Font */
 
 static void pango_ft2_font_class_init (PangoFT2FontClass *class);
-static void pango_ft2_font_init       (PangoFT2Font      *xfont);
+static void pango_ft2_font_init       (PangoFT2Font      *ft2font);
 static void pango_ft2_font_dispose    (GObject         *object);
 static void pango_ft2_font_finalize   (GObject         *object);
 
@@ -147,16 +152,56 @@ pango_ft2_font_get_face (PangoFont      *font)
     
       error = FT_New_Face (_pango_ft2_font_map_get_library (ft2font->fontmap),
 			   filename, id, &ft2font->face);
+      if (error)
+	{
+	  MiniXftPattern *sans;
+	  MiniXftPattern *matched;
+	  MiniXftResult result;
+	  char *filename2 = NULL;
+	  gchar *name;
+
+	bail0:
+	  
+	  sans = MiniXftPatternBuild (0,
+				      XFT_FAMILY, MiniXftTypeString, "sans",
+				      XFT_ENCODING, MiniXftTypeString, "glyphs-fontspecific",
+				      XFT_SIZE, MiniXftTypeDouble, (double)pango_font_description_get_size (ft2font->description)/PANGO_SCALE,
+				      NULL);
+
+	  matched = MiniXftFontMatch ((Display *)1, 0, sans, &result);
+
+	  if (MiniXftPatternGetString (matched, XFT_FILE, 0, &filename2) != MiniXftResultMatch)
+	    goto bail1;
+	  
+	  if (MiniXftPatternGetInteger (matched, XFT_INDEX, 0, &id) != MiniXftResultMatch)
+	    goto bail1;
+	  
+	  error = FT_New_Face (_pango_ft2_font_map_get_library (ft2font->fontmap),
+			       filename2, id, &ft2font->face);
+
+	  
+	  if (error)
+	    {
+	    bail1:
+	      name = pango_font_description_to_string (ft2font->description);
+	      g_warning ("Unable to open font file %s for font %s, exiting\n", filename2, name);
+	      exit (1);
+	    }
+	  else
+	    {
+	      name = pango_font_description_to_string (ft2font->description);
+	      g_warning ("Unable to open font file %s for font %s, falling back to %s\n", filename, name, filename2);
+	      g_free (name);
+	    }
+
+	  MiniXftPatternDestroy (sans);
+	  MiniXftPatternDestroy (matched);
+	}
       ft2font->face->generic.data = 0;
     }
- bail0:
-  
-  if (!ft2font->face)
-    {
-      g_warning ("Cannot load font\n");
-      return NULL;
-    }
 
+  g_assert (ft2font->face);
+  
   face = ft2font->face;
 
   if (ft2font->size != GPOINTER_TO_UINT (face->generic.data))
@@ -242,6 +287,8 @@ pango_ft2_font_init (PangoFT2Font *ft2font)
 
   ft2font->size = 0;
 
+  ft2font->metrics_by_lang = NULL;
+
   ft2font->glyph_info = g_hash_table_new (NULL, NULL);
 }
 
@@ -537,41 +584,108 @@ pango_ft2_font_get_kerning (PangoFont *font,
 }
 
 static PangoFontMetrics *
-pango_ft2_font_get_metrics (PangoFont        *font,
-			    PangoLanguage    *language)
+pango_ft2_font_get_metrics (PangoFont     *font,
+			    PangoLanguage *language)
 {
-  PangoFontMetrics *metrics;
-  FT_Face face;
-  
-  face = pango_ft2_font_get_face (font);
+  PangoFT2Font *ft2font = PANGO_FT2_FONT (font);
+  PangoFT2MetricsInfo *info = NULL; /* Quiet gcc */
+  GSList *tmp_list;      
 
-  metrics = pango_font_metrics_new ();
+  const char *sample_str = pango_language_get_sample_string (language);
   
-  metrics->ascent = PANGO_UNITS_26_6 (face->size->metrics.ascender);
-  metrics->descent = PANGO_UNITS_26_6 (-face->size->metrics.descender);
-  metrics->approximate_digit_width = PANGO_UNITS_26_6 (face->size->metrics.max_advance);
-  metrics->approximate_char_width = PANGO_UNITS_26_6 (face->size->metrics.max_advance);
+  tmp_list = ft2font->metrics_by_lang;
+  while (tmp_list)
+    {
+      info = tmp_list->data;
+      
+      if (info->sample_str == sample_str)    /* We _don't_ need strcmp */
+	break;
 
-  return pango_font_metrics_ref (metrics);
+      tmp_list = tmp_list->next;
+    }
+
+  if (!tmp_list)
+    {
+      PangoContext *context;
+      PangoLayout *layout;
+      PangoRectangle extents;
+      FT_Face face = pango_ft2_font_get_face (font);
+
+      info = g_new (PangoFT2MetricsInfo, 1);
+      info->sample_str = sample_str;
+      info->metrics = pango_font_metrics_new ();
+
+      info->metrics->ascent = PANGO_UNITS_26_6 (face->size->metrics.ascender);
+      info->metrics->descent = PANGO_UNITS_26_6 (- face->size->metrics.descender);
+      info->metrics->approximate_char_width = 
+        info->metrics->approximate_digit_width = 
+        PANGO_UNITS_26_6 (face->size->metrics.max_advance);
+
+      ft2font->metrics_by_lang = g_slist_prepend (ft2font->metrics_by_lang, info);
+
+      context = pango_context_new ();
+      pango_context_set_font_map (context, ft2font->fontmap);
+      pango_context_set_language (context, language);
+      
+      layout = pango_layout_new (context);
+      pango_layout_set_font_description (layout, ft2font->description);
+
+      pango_layout_set_text (layout, sample_str, -1);      
+      pango_layout_get_extents (layout, NULL, &extents);
+      
+      info->metrics->approximate_char_width = 
+        extents.width / g_utf8_strlen (sample_str, -1);
+
+      pango_layout_set_text (layout, "0123456789", -1);      
+      pango_layout_get_extents (layout, NULL, &extents);
+      
+      info->metrics->approximate_digit_width = extents.width / 10;
+
+      g_object_unref (G_OBJECT (layout));
+      g_object_unref (G_OBJECT (context));
+    }
+
+  return pango_font_metrics_ref (info->metrics);
 }
 
 static PangoCoverage *
 pango_ft2_calc_coverage (PangoFont     *font,
 			 PangoLanguage *language)
 {
-  PangoCoverage *result;
+  PangoCoverage *coverage;
   FT_Face face;
-  gunichar wc;
 
-  result = pango_coverage_new ();
+  coverage = pango_coverage_new ();
   face = pango_ft2_font_get_face (font);
-  for (wc = 0; wc < 65536; wc++)
-    {
-      if (FT_Get_Char_Index (face, wc))
-	pango_coverage_set (result, wc, PANGO_COVERAGE_EXACT);
-    }
 
-  return result;
+#ifdef HAVE_FT_GET_FIRST_CHAR
+  {
+    FT_UInt gindex;
+    FT_ULong charcode;
+    
+    charcode = FT_Get_First_Char (face, &gindex);
+    while (gindex)
+      {
+	pango_coverage_set (coverage, charcode, PANGO_COVERAGE_EXACT);
+	charcode = FT_Get_Next_Char (face, charcode, &gindex);
+      }
+  }
+#else  
+  /* Ugh, this is going to be SLOW */
+  {
+    gunichar wc;
+    
+    for (wc = 0; wc < G_MAXUSHORT; wc++)
+      {
+	FT_UInt glyph = FT_Get_Char_Index (face, wc);
+	
+	if (glyph && glyph < face->num_glyphs)
+	  pango_coverage_set (coverage, wc, PANGO_COVERAGE_EXACT);
+      }
+  }
+#endif
+
+  return coverage;
 }
 
 static void
@@ -603,6 +717,13 @@ pango_ft2_free_glyph_info_callback (gpointer key, gpointer value, gpointer data)
   return TRUE;
 }
 
+static void
+free_metrics_info (PangoFT2MetricsInfo *info)
+{
+  pango_font_metrics_unref (info->metrics);
+  g_free (info);
+}
+
 static void
 pango_ft2_font_finalize (GObject *object)
 {
@@ -621,6 +742,9 @@ pango_ft2_font_finalize (GObject *object)
   
   g_object_unref (G_OBJECT (ft2font->fontmap));
 
+  g_slist_foreach (ft2font->metrics_by_lang, (GFunc)free_metrics_info, NULL);
+  g_slist_free (ft2font->metrics_by_lang);  
+
   g_hash_table_foreach_remove (ft2font->glyph_info,
 			       pango_ft2_free_glyph_info_callback, object);
   g_hash_table_destroy (ft2font->glyph_info);
