@@ -1,5 +1,5 @@
 /* Pango
- * pangowin32-fontmap.c: Font handling
+ * pangowin32-fontmap.c: Win32 font handling
  *
  * Copyright (C) 2000 Red Hat Software
  * Copyright (C) 2000 Tor Lillqvist
@@ -101,6 +101,7 @@ static void       pango_win32_font_map_class_init    (PangoFontMapClass       *c
 
 static void       pango_win32_font_map_finalize      (GObject                      *object);
 static PangoFont *pango_win32_font_map_load_font     (PangoFontMap                 *fontmap,
+						      PangoContext                 *context,
 						      const PangoFontDescription   *description);
 static void       pango_win32_font_map_list_families (PangoFontMap                 *fontmap,
 						       PangoFontFamily            ***families,
@@ -154,10 +155,12 @@ static guint
 logfont_nosize_hash (gconstpointer v)
 {
   const LOGFONT *lfp = v;
+  gchar *face = g_ascii_strdown (lfp->lfFaceName, -1);
+  guint result = g_str_hash (face) + lfp->lfItalic + lfp->lfWeight;
 
-  return g_str_hash (lfp->lfFaceName) +
-    lfp->lfItalic +
-    lfp->lfWeight;
+  g_free (face);
+
+  return result;
 }
 
 /* Ditto comparison function */
@@ -167,7 +170,7 @@ logfont_nosize_equal (gconstpointer v1,
 {
   const LOGFONT *lfp1 = v1, *lfp2 = v2;
 
-  return (strcmp (lfp1->lfFaceName, lfp2->lfFaceName) == 0
+  return (g_ascii_strcasecmp (lfp1->lfFaceName, lfp2->lfFaceName) == 0
 	  && lfp1->lfItalic == lfp2->lfItalic
 	  && lfp1->lfWeight == lfp2->lfWeight);
 }
@@ -223,11 +226,20 @@ pango_win32_enum_proc (LOGFONT    *lfp,
 
   lf = *lfp;
 
-  EnumFontFamiliesEx (pango_win32_hdc, &lf, (FONTENUMPROC) pango_win32_inner_enum_proc, lParam, 0);
+  EnumFontFamiliesExA (pango_win32_hdc, &lf, (FONTENUMPROC) pango_win32_inner_enum_proc, lParam, 0);
 
   return 1;
 }
 
+/**
+ * pango_win32_font_map_for_display:
+ *
+ * Returns a #PangoWin32FontMap. Font maps are cached and should
+ * not be freed. If the font map is no longer needed, it can
+ * be released with pango_win32_shutdown_display().
+ *
+ * Returns: a #PangoFontMap.
+ **/
 PangoFontMap *
 pango_win32_font_map_for_display (void)
 {
@@ -247,7 +259,7 @@ pango_win32_font_map_for_display (void)
 
   memset (&logfont, 0, sizeof (logfont));
   logfont.lfCharSet = DEFAULT_CHARSET;
-  EnumFontFamiliesEx (pango_win32_hdc, &logfont, (FONTENUMPROC) pango_win32_enum_proc, 0, 0);
+  EnumFontFamiliesExA (pango_win32_hdc, &logfont, (FONTENUMPROC) pango_win32_enum_proc, 0, 0);
 
 #ifdef _WE_WANT_GLOBAL_ALIASES_
   pango_win32_font_map_read_aliases (fontmap);
@@ -419,6 +431,7 @@ pango_win32_get_font_family (PangoWin32FontMap *win32fontmap,
 
 static PangoFont *
 pango_win32_font_map_load_font (PangoFontMap               *fontmap,
+				PangoContext               *context,
 				const PangoFontDescription *description)
 {
   PangoWin32FontMap *win32fontmap = (PangoWin32FontMap *)fontmap;
@@ -431,6 +444,8 @@ pango_win32_font_map_load_font (PangoFontMap               *fontmap,
   
   name = g_ascii_strdown (pango_font_description_get_family (description), -1);
 
+  PING(("name=%s", name));
+
   win32family = g_hash_table_lookup (win32fontmap->families, name);
   if (win32family)
     {
@@ -454,6 +469,8 @@ pango_win32_font_map_load_font (PangoFontMap               *fontmap,
 	  GSList *tmp_list = best_match->cached_fonts;
 	  gint size = pango_font_description_get_size (description);
 
+	  PING(("got best match:%s",best_match->logfont.lfFaceName));
+
 	  while (tmp_list)
 	    {
 	      PangoWin32Font *win32font = tmp_list->data;
@@ -474,6 +491,7 @@ pango_win32_font_map_load_font (PangoFontMap               *fontmap,
 	    {
 	      PangoWin32Font *win32font;
 	      
+	      PING((""));
 	      win32font = pango_win32_font_new (fontmap, &best_match->logfont, size);
 	      win32font->fontmap = fontmap;
 	      win32font->entry = best_match;
@@ -709,14 +727,17 @@ pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
   PangoVariant variant;
   PangoWeight weight;
   PangoStretch stretch;
+  int i;
 
-  PING(("lfp.face=%s,wt=%ld,ht=%ld",lfp->lfFaceName,lfp->lfWeight,lfp->lfHeight));
+  PING(("lfp.face=%s,charset=%d,it=%d,wt=%ld,ht=%ld",lfp->lfFaceName,lfp->lfCharSet,lfp->lfItalic,lfp->lfWeight,lfp->lfHeight));
   
   /* First insert the LOGFONT into the list of LOGFONTs for the typeface name
    */
   lfp2 = g_new (LOGFONT, 1);
   *lfp2 = *lfp;
-  g_strdown (lfp2->lfFaceName);
+  /* lfp2->lfFaceName is an array of CHAR, change inplace */
+  for (i = 0; i < LF_FACESIZE; i++)
+    lfp2->lfFaceName[i] = g_ascii_tolower (lfp2->lfFaceName[i]);
   size_info = g_hash_table_lookup (win32fontmap->size_infos, lfp);
   if (!size_info)
     {
@@ -731,7 +752,7 @@ pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
   
   /* Convert the LOGFONT into a PangoFontDescription */
   
-  family_name = g_strdup (lfp2->lfFaceName);
+  family_name = lfp2->lfFaceName;
   
   if (!lfp2->lfItalic)
     style = PANGO_STYLE_NORMAL;
@@ -794,7 +815,8 @@ pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
   win32face->coverage = NULL;
   win32face->logfont = *lfp;
   win32face->unicode_table = NULL;
-  g_strdown (win32face->logfont.lfFaceName);
+  for (i = 0; i < LF_FACESIZE; i++)
+    win32face->logfont.lfFaceName[i] = g_ascii_tolower (win32face->logfont.lfFaceName[i]);
   font_family->font_entries = g_slist_append (font_family->font_entries, win32face);
   win32fontmap->n_fonts++;
 
@@ -825,6 +847,15 @@ pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
       break;
     }
 
+  /* Some other magic names */
+
+  /* Recognize just "courier" for "courier new" */
+  if (g_ascii_strcasecmp (win32face->logfont.lfFaceName, "courier new") == 0)
+    {
+      font_family = pango_win32_get_font_family (win32fontmap, "courier");
+      font_family->font_entries = g_slist_append (font_family->font_entries, win32face);
+      win32fontmap->n_fonts++;
+    }
 }
 
 /* Given a LOGFONT and size, make a matching LOGFONT corresponding to
@@ -975,6 +1006,14 @@ pango_win32_font_entry_remove (PangoWin32Face *face,
   face->cached_fonts = g_slist_remove (face->cached_fonts, font);
 }
 
+/**
+ * pango_win32_font_map_get_font_cache:
+ * @font_map: a #PangoWin32FontMap.
+ *
+ * Obtains the font cache associated with the given font map.
+ *
+ * Returns: the #PangoWin32FontCache of @font_map.
+ **/
 PangoWin32FontCache *
 pango_win32_font_map_get_font_cache (PangoFontMap *font_map)
 {
