@@ -1,5 +1,5 @@
 /* Pango
- * pangoxft-fontmap.h: Font handling
+ * pangoxft-fontmap.c: Xft font handling
  *
  * Copyright (C) 2000 Red Hat Software
  *
@@ -19,19 +19,31 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <string.h>
+
 #include "pango-fontmap.h"
 #include "pangoxft.h"
 #include "pangoxft-private.h"
 #include "modules.h"
 
+#include "X11/Xft/Xft.h"
 #include "X11/Xft/XftFreetype.h"
 
+#ifndef XFT_MINSPACE
+#define XFT_MINSPACE        "minspace"  /* Bool use minimum line spacing */
+#endif
+
+#ifndef XFT_MATRIX
+#define XFT_MATRIX          "matrix"    /* XftMatrix */
+#endif
+
 /* Number of freed fonts */
 #define MAX_FREED_FONTS 16
 
 typedef struct _PangoXftFontMap      PangoXftFontMap;
 typedef struct _PangoXftFamily       PangoXftFamily;
 typedef struct _PangoXftFace         PangoXftFace;
+typedef struct _PangoXftPatternSet   PangoXftPatternSet;
 
 #define PANGO_TYPE_XFT_FONT_MAP              (pango_xft_font_map_get_type ())
 #define PANGO_XFT_FONT_MAP(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_TYPE_XFT_FONT_MAP, PangoXftFontMap))
@@ -41,13 +53,19 @@ struct _PangoXftFontMap
 {
   PangoFontMap parent_instance;
 
-  GHashTable *font_hash;
-  GHashTable *coverage_hash;
-  GQueue *freed_fonts;
+  GHashTable *fontset_hash; /* Maps PangoFontDescription -> PangoXftPatternSet */
+  GHashTable *coverage_hash; /* Maps font file name -> PangoCoverage */
 
+  GHashTable *fonts; /* Maps XftPattern -> PangoXftFont */
+  GQueue *freed_fonts; /* Fonts in fonts that has been freed */
+
+  /* List of all families availible */
   PangoXftFamily **families;
   int n_families;		/* -1 == uninitialized */
 
+  /* List of all fonts (XftPatterns) availible */
+  XftFontSet *font_set;
+
   Display *display;
   int screen;  
 };
@@ -67,6 +85,12 @@ struct _PangoXftFamily
   int n_faces;		/* -1 == uninitialized */
 };
 
+struct _PangoXftPatternSet
+{
+  int n_patterns;
+  XftPattern **patterns;
+};
+
 #define PANGO_XFT_TYPE_FACE              (pango_xft_face_get_type ())
 #define PANGO_XFT_FACE(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_XFT_TYPE_FACE, PangoXftFace))
 #define PANGO_XFT_IS_FACE(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), PANGO_XFT_TYPE_FACE))
@@ -83,18 +107,26 @@ static GType    pango_xft_font_map_get_type   (void);
 GType           pango_xft_family_get_type     (void);
 GType           pango_xft_face_get_type       (void);
 
-static void       pango_xft_font_map_init          (PangoXftFontMap              *fontmap);
-static void       pango_xft_font_map_class_init    (PangoFontMapClass            *class);
-static void       pango_xft_font_map_finalize      (GObject                      *object);
-static PangoFont *pango_xft_font_map_load_font     (PangoFontMap                 *fontmap,
-						    const PangoFontDescription   *description);
-static void       pango_xft_font_map_list_families (PangoFontMap                 *fontmap,
-						    PangoFontFamily            ***families,
-						    int                          *n_families);
+static void          pango_xft_font_map_init          (PangoXftFontMap              *fontmap);
+static void          pango_xft_font_map_class_init    (PangoFontMapClass            *class);
+static void          pango_xft_font_map_finalize      (GObject                      *object);
+static PangoFont *   pango_xft_font_map_load_font     (PangoFontMap                 *fontmap,
+						       PangoContext                 *context,
+						       const PangoFontDescription   *description);
+static PangoFontset *pango_xft_font_map_load_fontset  (PangoFontMap                 *fontmap,
+						       PangoContext                 *context,
+						       const PangoFontDescription   *desc,
+						       PangoLanguage                *language);
+static void          pango_xft_font_map_list_families (PangoFontMap                 *fontmap,
+						       PangoFontFamily            ***families,
+						       int                          *n_families);
 
-static void pango_xft_font_map_cache_clear  (PangoXftFontMap *xfontmap);
-static void pango_xft_font_map_cache_remove (PangoFontMap    *fontmap,
-					     PangoXftFont    *xfont);
+
+static void pango_xft_font_set_free         (PangoXftPatternSet *font_set);
+
+static void pango_xft_font_map_cache_clear  (PangoXftFontMap    *xfontmap);
+static void pango_xft_font_map_cache_remove (PangoFontMap       *fontmap,
+					     PangoXftFont       *xfont);
 
 static PangoFontClass *parent_class;	/* Parent class structure for PangoXftFontMap */
 
@@ -141,11 +173,164 @@ pango_xft_font_map_class_init (PangoFontMapClass *class)
   
   object_class->finalize = pango_xft_font_map_finalize;
   class->load_font = pango_xft_font_map_load_font;
+  class->load_fontset = pango_xft_font_map_load_fontset;
   class->list_families = pango_xft_font_map_list_families;
 }
 
 static GSList *fontmaps = NULL;
 
+guint
+pango_xft_pattern_hash (XftPattern *pattern)
+{
+  char *str;
+  int i;
+  double d;
+  guint hash = 0;
+  
+  XftPatternGetString (pattern, XFT_FILE, 0, &str);
+  if (str)
+    hash = g_str_hash (str);
+
+  if (XftPatternGetInteger (pattern, XFT_INDEX, 0, &i) == XftResultMatch)
+    hash ^= i;
+
+  if (XftPatternGetDouble (pattern, XFT_PIXEL_SIZE, 0, &d) == XftResultMatch)
+    hash ^= (guint) (d*1000.0);
+
+  return hash;
+}
+
+typedef enum {
+  PATTERN_FILE       = 1 << 0,
+  PATTERN_INDEX      = 1 << 1,
+  PATTERN_PIXEL_SIZE = 1 << 3,
+  PATTERN_RGBA       = 1 << 4,
+  PATTERN_ANTIALIAS  = 1 << 5,
+  PATTERN_MINSPACE   = 1 << 6,
+  PATTERN_SPACING    = 1 << 7,
+  PATTERN_CHAR_WIDTH = 1 << 8
+} PatternElement;
+
+typedef struct {
+  char  *file;
+  int    index;
+  double pixel_size;
+  int    rgba;
+  Bool   antialias;
+  Bool   minspace;
+  int    spacing;
+  int    char_width;
+} PatternInfo;
+
+static PatternElement
+get_pattern_info (XftPattern  *pattern,
+		  PatternInfo *info)
+{
+  int i;
+  PatternElement fields = 0;
+  
+  for (i = 0; i < pattern->num; i++)
+    {
+      const char *object = pattern->elts[i].object;
+      XftValue *value;
+
+      if (!pattern->elts[i].values)
+	continue;
+
+      value = &pattern->elts[i].values->value;
+
+      switch (object[0])
+	{
+	case 'a':
+	  if (strcmp (object, XFT_ANTIALIAS) == 0 && value->type == XftTypeBool)
+	    {
+	      info->antialias = value->u.b;
+	      fields |= PATTERN_ANTIALIAS;
+	    }
+	  break;
+	case 'c':
+	  if (strcmp (object, XFT_CHAR_WIDTH) == 0 && value->type == XftTypeInteger)
+	    {
+	      info->char_width = value->u.i;
+	      fields |= PATTERN_CHAR_WIDTH;
+	    }
+	  break;
+	case 'f':
+	  if (strcmp (object, XFT_FILE) == 0 && value->type == XftTypeString)
+	    {
+	      info->file = value->u.s;
+	      fields |= PATTERN_FILE;
+	    }
+	  break;
+	case 'i':
+	  if (strcmp (object, XFT_INDEX) == 0 && value->type == XftTypeInteger)
+	    {
+	      info->index = value->u.i;
+	      fields |= PATTERN_INDEX;
+	    }
+	  break;
+	case 'm':
+	  if (strcmp (object, XFT_MINSPACE) == 0 && value->type == XftTypeBool)
+	    {
+	      info->minspace = value->u.b;
+	      fields |= PATTERN_MINSPACE;
+	    }
+	  break;
+	case 'p':
+	  if (strcmp (object, XFT_PIXEL_SIZE) == 0 && value->type == XftTypeDouble)
+	    {
+	      info->pixel_size = value->u.d;
+	      fields |= PATTERN_PIXEL_SIZE;
+	      break;
+	    }
+	case 'r':
+	  if (strcmp (object, XFT_RGBA) == 0 && value->type == XftTypeInteger)
+	    {
+	      info->rgba = value->u.i;
+	      fields |= PATTERN_RGBA;
+	    }
+	  break;
+	case 's':
+	  if (strcmp (object, XFT_SPACING) == 0 && value->type == XftTypeInteger)
+	    {
+	      info->spacing = value->u.i;
+	      fields |= PATTERN_SPACING;
+	      break;
+	    }
+	  break;
+	}
+    }
+
+  return fields;
+}
+
+gboolean
+pango_xft_pattern_equal (XftPattern *pattern1,
+			 XftPattern *pattern2)
+{
+  PatternInfo info1;
+  PatternInfo info2;
+  PatternElement elements1, elements2;
+
+  elements1 = get_pattern_info (pattern1, &info1);
+  elements2 = get_pattern_info (pattern2, &info2);
+
+  if (elements1 != elements2)
+    return FALSE;
+
+  if (((elements1 & PATTERN_FILE) &&       strcmp (info1.file, info2.file) != 0) ||
+      ((elements1 & PATTERN_INDEX) &&      info1.index != info2.index) ||
+      ((elements1 & PATTERN_PIXEL_SIZE) && info1.pixel_size != info2.pixel_size) ||
+      ((elements1 & PATTERN_RGBA) &&       info1.rgba != info2.rgba) ||
+      ((elements1 & PATTERN_ANTIALIAS) &&  info1.antialias != info2.antialias) ||
+      ((elements1 & PATTERN_MINSPACE) &&   info1.minspace != info2.minspace) ||
+      ((elements1 & PATTERN_SPACING) &&    info1.spacing != info2.spacing) ||
+      ((elements1 & PATTERN_CHAR_WIDTH) && info1.char_width != info2.char_width))
+    return FALSE;
+  
+  return TRUE;
+}
+
 static PangoFontMap *
 pango_xft_get_font_map (Display *display,
 			int      screen)
@@ -165,6 +350,8 @@ pango_xft_get_font_map (Display *display,
       if (xfontmap->display == display &&
 	  xfontmap->screen == screen) 
 	return PANGO_FONT_MAP (xfontmap);
+
+      tmp_list = tmp_list->next;
     }
 
   xfontmap = (PangoXftFontMap *)g_object_new (PANGO_TYPE_XFT_FONT_MAP, NULL);
@@ -172,9 +359,15 @@ pango_xft_get_font_map (Display *display,
   xfontmap->display = display;
   xfontmap->screen = screen;
 
-  xfontmap->font_hash = g_hash_table_new ((GHashFunc)pango_font_description_hash,
-					  (GEqualFunc)pango_font_description_equal);
-  xfontmap->coverage_hash = g_hash_table_new (g_str_hash, g_str_equal);
+  xfontmap->fonts = g_hash_table_new ((GHashFunc)pango_xft_pattern_hash,
+				      (GEqualFunc)pango_xft_pattern_equal);
+  xfontmap->fontset_hash = g_hash_table_new_full ((GHashFunc)pango_font_description_hash,
+						  (GEqualFunc)pango_font_description_equal,
+						  (GDestroyNotify)pango_font_description_free,
+						  (GDestroyNotify)pango_xft_font_set_free);
+  xfontmap->coverage_hash = g_hash_table_new_full (g_str_hash, g_str_equal,
+						   (GDestroyNotify)g_free,
+						   (GDestroyNotify)pango_coverage_unref);
   xfontmap->freed_fonts = g_queue_new ();
 
   fontmaps = g_slist_prepend (fontmaps, xfontmap);
@@ -211,20 +404,11 @@ pango_xft_get_context (Display *display,
     }
   
   result = pango_context_new ();
-  pango_context_add_font_map (result, pango_xft_get_font_map (display, screen));
+  pango_context_set_font_map (result, pango_xft_get_font_map (display, screen));
 
   return result;
 }
 
-static void
-coverage_foreach (gpointer key, gpointer value, gpointer data)
-{
-  PangoCoverage *coverage = value;
-
-  g_free (key);
-  pango_coverage_unref (coverage);
-}
-
 static void
 pango_xft_font_map_finalize (GObject *object)
 {
@@ -233,30 +417,34 @@ pango_xft_font_map_finalize (GObject *object)
   fontmaps = g_slist_remove (fontmaps, object);
 
   g_queue_free (xfontmap->freed_fonts);
-  g_hash_table_destroy (xfontmap->font_hash);
-
-  g_hash_table_foreach (xfontmap->coverage_hash, coverage_foreach, NULL);
+  g_hash_table_destroy (xfontmap->fontset_hash);
   g_hash_table_destroy (xfontmap->coverage_hash);
   
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+
+/* Add a mapping from xfont->font_pattern to xfont */
 void
 _pango_xft_font_map_add (PangoFontMap *fontmap,
-			PangoXftFont *xfont)
+			 PangoXftFont *xfont)
 {
   PangoXftFontMap *xfontmap = PANGO_XFT_FONT_MAP (fontmap);
 
-  g_hash_table_insert (xfontmap->font_hash, xfont->description, xfont);
+  g_hash_table_insert (xfontmap->fonts,
+		       xfont->font_pattern,
+		       xfont);
 }
 
+/* Remove mapping from xfont->font_pattern to xfont */
 void
 _pango_xft_font_map_remove (PangoFontMap *fontmap,
 			   PangoXftFont *xfont)
 {
   PangoXftFontMap *xfontmap = PANGO_XFT_FONT_MAP (fontmap);
 
-  g_hash_table_remove (xfontmap->font_hash, xfont->description);
+  g_hash_table_remove (xfontmap->fonts,
+		       xfont->font_pattern);
 }
 
 static void
@@ -303,38 +491,10 @@ pango_xft_font_map_list_families (PangoFontMap           *fontmap,
     *families = g_memdup (xfontmap->families, xfontmap->n_families * sizeof (PangoFontFamily *));
 }
 
-static PangoFont *
-pango_xft_font_map_load_font (PangoFontMap               *fontmap,
-			      const PangoFontDescription *description)
+static int
+pango_xft_convert_weight (PangoWeight pango_weight)
 {
-  PangoXftFontMap *xfontmap = (PangoXftFontMap *)fontmap;
-  PangoXftFont *font;
-  PangoStyle pango_style;
-  int slant;
-  PangoWeight pango_weight;
   int weight;
-  XftFont *xft_font;
-
-  font = g_hash_table_lookup (xfontmap->font_hash, description);
-
-  if (font)
-    {
-      if (font->in_cache)
-	pango_xft_font_map_cache_remove (fontmap, font);
-      
-      return (PangoFont *)g_object_ref (G_OBJECT (font));
-    }
-
-  pango_style = pango_font_description_get_style (description);
-  
-  if (pango_style == PANGO_STYLE_ITALIC)
-    slant = XFT_SLANT_ITALIC;
-  else if (pango_style == PANGO_STYLE_OBLIQUE)
-    slant = XFT_SLANT_OBLIQUE;
-  else
-    slant = XFT_SLANT_ROMAN;
-
-  pango_weight = pango_font_description_get_weight (description);
   
   if (pango_weight < (PANGO_WEIGHT_NORMAL + PANGO_WEIGHT_LIGHT) / 2)
     weight = XFT_WEIGHT_LIGHT;
@@ -346,55 +506,216 @@ pango_xft_font_map_load_font (PangoFontMap               *fontmap,
     weight = XFT_WEIGHT_BOLD;
   else
     weight = XFT_WEIGHT_BLACK;
+  
+  return weight;
+}
 
+static int
+pango_xft_convert_slant (PangoStyle pango_style)
+{
+  int slant;
+  
+  if (pango_style == PANGO_STYLE_ITALIC)
+    slant = XFT_SLANT_ITALIC;
+  else if (pango_style == PANGO_STYLE_OBLIQUE)
+    slant = XFT_SLANT_OBLIQUE;
+  else
+    slant = XFT_SLANT_ROMAN;
+  
+  return slant;
+}
+
+
+static XftPattern *
+pango_xft_make_pattern (const PangoFontDescription *description)
+{
+  XftPattern *pattern;
+  PangoStyle pango_style;
+  int slant;
+  int weight;
+  char **families;
+  int i;
+
+  pango_style = pango_font_description_get_style (description);
+
+  slant = pango_xft_convert_slant (pango_style);
+  weight = pango_xft_convert_weight (pango_font_description_get_weight (description));
+  
   /* To fool Xft into not munging glyph indices, we open it as glyphs-fontspecific
    * then set the encoding ourself
    */
-  xft_font = XftFontOpen (xfontmap->display, xfontmap->screen,
-			  XFT_ENCODING, XftTypeString, "glyphs-fontspecific",
-			  XFT_CORE, XftTypeBool, False,
-			  XFT_FAMILY, XftTypeString,  pango_font_description_get_family (description),
-			  XFT_WEIGHT, XftTypeInteger, weight,
-			  XFT_SLANT,  XftTypeInteger, slant,
-			  XFT_SIZE, XftTypeDouble, (double)pango_font_description_get_size (description)/PANGO_SCALE,
-			  NULL);
+  pattern = XftPatternBuild (0,
+			     XFT_ENCODING, XftTypeString, "glyphs-fontspecific",
+			     XFT_CORE, XftTypeBool, False,
+			     XFT_WEIGHT, XftTypeInteger, weight,
+			     XFT_SLANT,  XftTypeInteger, slant,
+			     XFT_SIZE, XftTypeDouble, (double)pango_font_description_get_size (description)/PANGO_SCALE,
+			     NULL);
 
-  if (xft_font)
+  families = g_strsplit (pango_font_description_get_family (description), ",", -1);
+  
+  for (i = 0; families[i]; i++)
+    XftPatternAddString (pattern, XFT_FAMILY, families[i]);
+
+  g_strfreev (families);
+
+  return pattern;
+}
+
+static PangoFont *
+pango_xft_font_map_new_font (PangoFontMap  *fontmap,
+			     XftPattern    *match)
+{
+  PangoXftFontMap *xfontmap = (PangoXftFontMap *)fontmap;
+  PangoXftFont *font;
+  
+  /* Look up cache */
+  font = g_hash_table_lookup (xfontmap->fonts, match);
+  
+  if (font)
     {
-      FT_Face face;
-      FT_Error error;
-      
-      int charmap;
+      g_object_ref (font);
 
-      g_assert (!xft_font->core);
-      
-      face = xft_font->u.ft.font->face;
+      /* Revive font from cache */
+      if (font->in_cache)
+	pango_xft_font_map_cache_remove (fontmap, font);
 
-      for (charmap = 0; charmap < face->num_charmaps; charmap++)
-	if (face->charmaps[charmap]->encoding == ft_encoding_unicode)
-	  break;
-
-      if (charmap == face->num_charmaps)
-	goto error;
-
-      error = FT_Set_Charmap(face, face->charmaps[charmap]);
-      
-      if (error)
-	goto error;
-      
-      font = _pango_xft_font_new (fontmap, description, xft_font);
+      return (PangoFont *)font;
     }
-  else
-    return NULL;
+  
+  return  (PangoFont *)_pango_xft_font_new (fontmap, XftPatternDuplicate (match));
+}
 
-  return (PangoFont *)font;
+static PangoXftPatternSet *
+pango_xft_font_map_get_patterns (PangoFontMap               *fontmap,
+				 PangoContext               *context,
+				 const PangoFontDescription *desc)
+{
+  PangoXftFontMap *xfontmap = (PangoXftFontMap *)fontmap;
+  XftPattern *pattern, *pattern_copy;
+  XftPattern *match;
+  char *family, *family_res;
+  XftResult res;
+  int id;
+  GPtrArray *array;
+  PangoXftPatternSet *patterns;
 
- error:
+  patterns = g_hash_table_lookup (xfontmap->fontset_hash, desc);
 
-  XftFontClose (xfontmap->display, xft_font);
+  if (patterns == NULL)
+    {
+      if (xfontmap->font_set == NULL)
+	xfontmap->font_set = XftListFonts (xfontmap->display, xfontmap->screen,
+					   XFT_CORE, XftTypeBool, False,
+					   XFT_ENCODING, XftTypeString, "iso10646-1",
+					   NULL,
+					   XFT_FOUNDRY,    XFT_STYLE,    XFT_FAMILY,
+					   XFT_ENCODING,   XFT_FILE,     XFT_INDEX,
+					   XFT_CORE,       XFT_WEIGHT,   XFT_SLANT,
+					   XFT_CHAR_WIDTH, XFT_MATRIX,   XFT_RGBA,
+					   XFT_ANTIALIAS,  XFT_MINSPACE, XFT_SPACING,
+					   XFT_SIZE,
+					   NULL);
+      
+      pattern = pango_xft_make_pattern (desc);
+
+      XftConfigSubstitute (pattern);
+      XftDefaultSubstitute (xfontmap->display, xfontmap->screen, pattern);
+
+      pattern_copy = XftPatternDuplicate (pattern);
+
+      array = g_ptr_array_new ();
+      patterns = g_new (PangoXftPatternSet, 1);
+
+      match = NULL;
+      id = 0;
+      while (XftPatternGetString (pattern, XFT_FAMILY, id++, &family) == XftResultMatch)
+	{
+	  XftPatternDel (pattern_copy, XFT_FAMILY);
+	  XftPatternAddString (pattern_copy, XFT_FAMILY, family);
+	  
+	  match = XftFontSetMatch (&xfontmap->font_set, 1, pattern_copy, &res);
+	  
+	  if (match &&
+	      XftPatternGetString (match, XFT_FAMILY, 0, &family_res) == XftResultMatch &&
+	      g_ascii_strcasecmp (family, family_res) == 0)
+	    {
+	      g_ptr_array_add (array, match);
+	      match = NULL;
+	    }
+	  if (match)
+	    XftPatternDestroy (match);
+	}
+      
+      if (array->len == 0)
+	{
+	  match = XftFontSetMatch (&xfontmap->font_set, 1, pattern, &res);
+	  if (match == NULL)
+	    g_error ("Failed to match any font. This could be due to a broken Xft "
+		     "configuration, or if you run XFree 4.1.0 due to a bug in libXrender. "
+		     "For more information about this, read http://bugzilla.gnome.org/show_bug.cgi?id=68030\n");
+	  g_ptr_array_add (array, match);
+	}
+
+      XftPatternDestroy (pattern);
+      XftPatternDestroy (pattern_copy);
+
+      patterns->n_patterns = array->len;
+      patterns->patterns = (XftPattern **)g_ptr_array_free (array, FALSE);
+      
+      g_hash_table_insert (xfontmap->fontset_hash,
+			   pango_font_description_copy (desc),
+			   patterns);
+    }
+
+  return patterns;
+}
+
+static PangoFont *
+pango_xft_font_map_load_font (PangoFontMap               *fontmap,
+			      PangoContext               *context,
+			      const PangoFontDescription *description)
+{
+  PangoXftPatternSet *patterns = pango_xft_font_map_get_patterns (fontmap, context, description);
+
+  if (patterns->n_patterns > 0)
+    return pango_xft_font_map_new_font (fontmap, patterns->patterns[0]);
+  
   return NULL;
 }
 
+static void
+pango_xft_font_set_free (PangoXftPatternSet *font_set)
+{
+  int i;
+  
+  for (i = 0; i < font_set->n_patterns; i++)
+    XftPatternDestroy (font_set->patterns[i]);
+
+  g_free (font_set);
+}
+
+
+static PangoFontset *
+pango_xft_font_map_load_fontset (PangoFontMap                 *fontmap,
+				 PangoContext                 *context,
+				 const PangoFontDescription   *desc,
+				 PangoLanguage                *language)
+{
+  PangoFontsetSimple *simple;
+  int i;
+  PangoXftPatternSet *patterns = pango_xft_font_map_get_patterns (fontmap, context, desc);
+	  
+  simple = pango_fontset_simple_new (language);
+
+  for (i = 0; i < patterns->n_patterns; i++)
+    pango_fontset_simple_append (simple,
+				 pango_xft_font_map_new_font (fontmap, patterns->patterns[i]));
+  
+  return PANGO_FONTSET (simple);
+}
+
+
 void
 _pango_xft_font_map_cache_add (PangoFontMap *fontmap,
 			       PangoXftFont *xfont)
@@ -482,12 +803,13 @@ _pango_xft_font_map_get_info (PangoFontMap *fontmap,
  * PangoXftFace
  */
 
-static PangoFontDescription *
-font_desc_from_pattern (XftPattern *pattern)
+PangoFontDescription *
+_pango_xft_font_desc_from_pattern (XftPattern *pattern, gboolean include_size)
 {
   PangoFontDescription *desc;
   PangoStyle style;
   PangoWeight weight;
+  double size;
   
   char *s;
   int i;
@@ -530,6 +852,9 @@ font_desc_from_pattern (XftPattern *pattern)
   else
     weight = PANGO_WEIGHT_NORMAL;
 
+  if (include_size && XftPatternGetDouble (pattern, XFT_SIZE, 0, &size) == XftResultMatch)
+      pango_font_description_set_size (desc, size * PANGO_SCALE);
+  
   pango_font_description_set_weight (desc, weight);
   
   pango_font_description_set_variant (desc, PANGO_VARIANT_NORMAL);
@@ -560,7 +885,7 @@ pango_xft_face_describe (PangoFontFace *face)
   result_pattern = XftFontMatch (xfontmap->display, xfontmap->screen, match_pattern, &res);
   if (result_pattern)
     {
-      desc = font_desc_from_pattern (result_pattern);
+      desc = _pango_xft_font_desc_from_pattern (result_pattern, FALSE);
       XftPatternDestroy (result_pattern);
     }
 
@@ -645,7 +970,8 @@ pango_xft_family_list_faces (PangoFontFamily  *family,
 	  XftResult res;
 	  
 	  res = XftPatternGetString (fontset->fonts[i], XFT_STYLE, 0, &s);
-	  g_assert (res == XftResultMatch);
+	  if (res != XftResultMatch)
+	    s = "Regular";
 	  
 	  xfamily->faces[i] = g_object_new (PANGO_XFT_TYPE_FACE, NULL);
 	  xfamily->faces[i]->style = g_strdup (s);
